import numpy as np


def interp(p, xp, fp):
    """This is just a straight wrapper for np.interp(). The only reason I've added this is 
    because I knew you'd look here for an interpolation function. You probably should use 
    np.interp().

    Args:
        p (array_like): The x-coordinates at which to evaluate the interpolated values.
        xp (1-D sequence of floats): The x-coordinates of the data points, must be increasing if argument period is not specified. Otherwise, xp is internally sorted after normalizing the periodic boundaries with xp = xp % period.
        fp (1-D sequence of float or complex): The y-coordinates of the data points, same length as xp.
    """
    answer = np.interp(p, xp, fp)
    return answer



import pandas as pd
import altair as alt
import folium


def simple_station_map(centroid_coords, area_geojson, station_df: pd.DataFrame, zoom_level=10, extra_geojson1=None, extra_geojson2=None):
    """Returns a map of stations with pop up of information provided for each station

    Args:
        centroid_coords (array of doubles): centroid coordinates for your map
        area_geojson (geojson layer): outline layer for your map (i.e., catchment, reach, state)
        station_df (pd.DataFrame): dataframe with station details. Must include 'Station name', 'Station number', 'Latitude' and 'Longitude'. Any other columns are fine and information will be included in map pop up.
        zoom_level (int, optional): zoom level for your map. Defaults to 10.
        extra_geojson1 (geojson layer, optional): additional mapping layer. Defaults to None.
        extra_geojson2 (geojson layer, optional): additional mapping layer. Defaults to None.

    Returns:
        folium map: map of stations
    """

    # Set up map
    # My preferred tiles are 'Stamen Terrain', but they are no longer available. Might be available through StadiaMap, but haven't figured that out yet.
    m = folium.Map(location=centroid_coords, zoom_start=zoom_level, tiles="OpenStreetMap")
    
    # Add additional geojson map layers if the user has specified them
    if extra_geojson2 is None:
        pass
    else:
        folium.GeoJson(extra_geojson2, 
                name='geojson', 
                style_function=lambda x: {'color': '#12239e', 'weight': 0.5}).add_to(m)

    if extra_geojson1 is None:
        pass
    else:
        folium.GeoJson(extra_geojson1, 
                name='geojson', 
                style_function=lambda x: {'color': '#12239e', 'weight': 0.5}).add_to(m)
    
    # Add area 
    folium.GeoJson(area_geojson, 
                name='geojson', 
                style_function=lambda x: {'color': '#b12ef7', 'weight': 3}).add_to(m)

    # Add markers for the stations
    for i in range(0,len(station_df)):
        row = station_df.iloc[i]
        number_name = f"{row['Station number']} ({row['Station name']})"
        location = [station_df.iloc[i]['Latitude'], station_df.iloc[i]['Longitude']]
        
        # Add information from station data frame into marker pop up
        stationInfo=''
        for c in range(0,len(station_df.columns)):
            stationInfo+=("<b>" + station_df.columns[c] + ":</b> " + f"{row[station_df.columns[c]]}<br>")
        stationPopup=folium.Popup(stationInfo,min_width=100,max_width=400)

        folium.Marker(
            location=location,
            popup=stationPopup,
            icon=folium.Icon(icon='circle', prefix='fa', color='blue'),
            tooltip=number_name
        ).add_to(m)
    return m


def station_map_plots(centroid_coords, area_geojson, station_df: pd.DataFrame, data_df, data_label='data', zoom_level=10):
    """Returns a map of markers with pop up of daily data plot

    Args:
        centroid_coords (array of doubles):centroid coordinates for your map
        area_geojson (geojson layer): outline layer for your map (i.e., catchment, reach, state)
        station_df (pd.DataFrame): dataframe with station details. Must include 'Name', 'Latitude' and 'Longitude'. Any other columns ignored.
        data_df (array of pd.DataFrame): daily data to plot in pop up
        data_label (str, optional): description of data, used in plot title. Defaults to 'data'.
        zoom_level (int, optional): zoom level for your map. Defaults to 10.

    Returns:
        _type_: map of stations with daily data plots
    """

    # Set up map
    # My preferred tiles are 'Stamen Terrain', but they are no longer available. Might be available through StadiaMap, but haven't figured that out yet.
    m = folium.Map(location=centroid_coords, zoom_start=zoom_level, tiles="OpenStreetMap")
    
    # Add area 
    folium.GeoJson(area_geojson, 
                name='geojson', 
                style_function=lambda x: {'color': '#b12ef7', 'weight': 3}).add_to(m)

    # Add markers for the stations
    for i in range(0,len(station_df)):
        row = station_df.iloc[i]
        name = f"{row['Name']}"
        location = [station_df.iloc[i]['Latitude'], station_df.iloc[i]['Longitude']]
        
        # Add daily data to plot and put in pop up
        alt.data_transformers.disable_max_rows()
        dataChart = alt.Chart(data_df[i]).mark_line().encode(
            alt.X('Date:T'),
            alt.Y(data_df[i].columns[1] + ':Q',
            axis=alt.Axis(title=data_label))
        ).properties(width=600,height=300,title=name).interactive()

        stationPopup=folium.Popup(min_width=100,max_width=700).add_child(folium.features.VegaLite(dataChart,height=350,width=700))

        folium.Marker(
            location=location,
            popup=stationPopup,
            icon=folium.Icon(icon='circle', prefix='fa', color='blue'),
            tooltip=name
        ).add_to(m)
    return m


import pandas as pd
import numpy as np
from odhpy import utils
from datetime import datetime, timedelta





def annual_max(df: pd.DataFrame, wy_month=7, allow_part_years=False):
    """Returns the maximum annual for a daily timeseries dataframe.

    Args:
        df (pd.DataFrame): Dataframe with date as index
        wy_month (int, optional): Water year start month. Defaults to 7.
        allow_part_years (bool, optional): Allow part water years or only complete water years. Defaults to False.

    Returns:
        _type_: _description_
    """
    if (allow_part_years):
        return df.groupby(utils.get_wy(df.index, wy_month)).sum().max()
    else:
        cropped_df = utils.crop_to_wy(df, wy_month)
        if (len(cropped_df) > 0):
            return cropped_df.groupby(utils.get_wy(cropped_df.index, wy_month)).sum().max()
        else:
            return np.nan

def annual_min(df: pd.DataFrame, wy_month=7, allow_part_years=False):
    """Returns the minimum annual for a daily timeseries dataframe.

    Args:
        df (pd.DataFrame): Dataframe with date as index
        wy_month (int, optional): Water year start month. Defaults to 7.
        allow_part_years (bool, optional): Allow part water years or only complete water years. Defaults to False.

    Returns:
        _type_: _description_
    """
    if (allow_part_years):
        return df.groupby(utils.get_wy(df.index, wy_month)).sum().max()
    else:
        cropped_df = utils.crop_to_wy(df, wy_month)
        if (len(cropped_df) > 0):
            return cropped_df.groupby(utils.get_wy(cropped_df.index, wy_month)).sum().min()
        else:
            return np.nan

def annual_mean(df: pd.DataFrame, wy_month=7, allow_part_years=False):
    """Returns the mean annual for a daily timeseries dataframe.

    Args:
        df (pd.DataFrame): Dataframe with date as index
        wy_month (int, optional): Water year start month. Defaults to 7.
        allow_part_years (bool, optional): Allow part water years or only complete water years. Defaults to False.

    Returns:
        _type_: _description_
    """
    if (allow_part_years):
        return df.groupby(utils.get_wy(df.index, wy_month)).sum().max()
    else:
        cropped_df = utils.crop_to_wy(df, wy_month)
        if (len(cropped_df) > 0):
            return cropped_df.groupby(utils.get_wy(cropped_df.index, wy_month)).sum().mean()
        else:
            return np.nan





import os
import numpy as np
import pandas as pd
from odhpy import utils
na_values = ['', ' ', 'null', 'NULL', 'NAN', 'NaN', 'nan', 'NA', 'na', 'N/A' 'n/a', '#N/A', '#NA', '-NaN', '-nan']



# def read_ts_csv(filename, date_format=None, df=None, colprefix="", **kwargs):
#     """Reads a daily timeseries csv into a DataFrame, and sets the index to the Date.
#     Assumed there is a column named "Date"

#     Args:
#         filename (_type_): _description_
#         date_format (str, optional): defaults to "%d/%m/%Y" as per Fors. Other common formats include "%Y-%m-%d", "%Y/%m/%d".

#     Returns:
#         _type_: _description_
#     """
#     # If no df was supplied, instantiate a new one
#     if df is None:
#         df = pd.DataFrame()
#     # Read the data
#     temp = pd.read_csv(filename, na_values=na_values, **kwargs)
#     temp = utils.set_index_dt(temp, dayfirst=True, format=date_format)    
#     temp.index.name = 'Date'
#     temp = temp.replace(r'^\s*$', np.nan, regex=True)
#     if colprefix is not None:
#         for c in temp.columns:
#             temp.rename(columns = {c:f"{colprefix}{c}"}, inplace = True)        
#     df = df.join(temp, how="outer").sort_index()
#     # TODO: THERE IS NO GUARANTEE THAT THE DATES OVERLAP, THEREFORE WE MAY END UP WITH A DATAFRAME WITH INCOMPLETE DATES
#     # TODO: I SHOULD MAKE DEFAULT BEHAVIOUR AUTO-DETECT FORMAT DEPENDING ON *TYPE* AND *LOCATION* OF DELIMIT CHARS
#     # TODO: In the meantime we use the below to assert that the format of the resulting df meets our minimum standards.
#     utils.assert_df_format_standards(df)
#     return df


def read_ts_csv(filename, date_format=None, df=None, colprefix=None, allow_nonnumeric=False, assert_date=True, **kwargs):
    """Reads a daily timeseries csv into a DataFrame, and sets the index to string dates in the "%Y-%m-%d" format. 
    The method assumes the first column are dates.

    Args:
        filename (_type_): _description_
        date_format (str, optional): defaults to "%d/%m/%Y" as per Fors. Other common formats include "%Y-%m-%d", "%Y/%m/%d".
        df (pd.DataFrame, optional): If provided, the reader will append columns to this dataframe. Defaults to None.
        colprefix (str, optional): If provided, the reader will append this prefix to the start of each column name. Defaults to None.
        allow_nonnumeric (bool, optional): If false, the method will assert that all columns are numerical. Defaults to False.
        assert_date (bool, optional): If true, the method will assert that date index meets "%Y-%m-%d" format. Defaults to True.         

    Returns:
        pd.DataFrame: Dataframe containing the data from the csv file.
    """
    new_df = pd.read_csv(filename, na_values=na_values, **kwargs)
    # Date index
    new_df.set_index(new_df.columns[0], inplace=True)
    if assert_date:
        new_df.index = utils.standardize_datestring_format(new_df.index)
    new_df.index.name = "Date"
    #df = df.replace(r'^\s*$', np.nan, regex=True)
    # Check values
    if not allow_nonnumeric:
        for col in new_df.columns:
            if not np.issubdtype(new_df[col].dtype, np.number):
                raise Exception(f"ERROR: Column '{col}' is not numeric!")
    # Rename columns if required
    if colprefix is not None:
        for c in new_df.columns:
            new_df.rename(columns = {c:f"{colprefix}{c}"}, inplace = True)
    # Join to existing dataframe if required
    if df is None:
        df = new_df
    else:
        if len(df) > 0:
            # Check that the dates overlap
            newdf_ends_before_df_starts = new_df.index[0] < df.index[-1]
            df_ends_before_newdf_starts = df.index[-1] < new_df.index[0]
            if newdf_ends_before_df_starts or df_ends_before_newdf_starts:
                raise Exception("ERROR: The dates in the new dataframe do not overlap with the existing dataframe!")
        df = df.join(new_df, how="outer")
    # Return
    return df


def write_ts_csv(df: pd.DataFrame, filename: str):
    """_summary_

    Args:
        df (pd.DataFrame): _description_
        filename (str): _description_
    """
    df.to_csv(filename)


import pandas as pd
import numpy as np
import calendar
from odhpy import utils
from datetime import datetime, timedelta
from typing import Union
from typing import Literal

class Reliability:

    def __init__(self, demand: Union[pd.Series,list,float,int], supply: pd.Series, demand_timescale: Literal["daily","monthly","yearly"]="daily", demand_type: Literal["total","daily_constant"]="total", ignore_leap_years=False) -> None:
        """Initialise Reliability class. Functions available are Reliability.DemandTS, Reliability.MonthlyReliability and Reliability.AnnualReliability.
        
        Args:
            demand (Union[pd.Series,list,float,int]): Demand timeseries with date string as index. Alternatively, a list of monthly values or single demand which will be disaggregated according to other input parameters.
            supply (pd.Series): Daily supply timeseries with date as index.
            demand_timescale (Literal["daily","monthly","yearly"], optional): If a float or int demand is provided, which timescale does it apply to. Defaults to "daily".
            demand_type (Literal["total","daily_constant"], optional): If a non-Series demand is provided, does the monthly list or float/int refer to a total demand over 'demand_timescale', or a daily constant value to be applied. Defaults to "total".
            ignore_leap_years (bool, optional): Decide whether leap years should be ignored in demand disaggregation. When True, calculations will always be based on 28 days in Feb, and value for the 29th Feb will be equal to the 28th of Feb value. Defaults to False.

        """        
        
        if type(demand) not in [pd.Series,list,float,int]:
            raise Exception("Demand must be one of timeseries, monthly list or single demand.")
        
        if type(supply) != pd.Series:
            raise Exception("Supply must be a single column of a date-indexed dataframe (pd.Series).")
        
        if type(demand) == pd.Series:
            print(f"Comparing provided demand timeseries with supply timeseries.")
            state = "ts"

        if type(demand) != pd.Series:
            if demand_type not in ["total","daily_constant"]:
                raise Exception("demand_type must be one of \"total\" or \"daily_constant\".")
            
            if demand_timescale not in ["daily","monthly","yearly"]:
                raise Exception("demand_timescale must be one of \"daily\", \"monthly\" or \"yearly\".")
            
            if ignore_leap_years not in [True, False]:
                    raise Exception(f"ignore_leap_years must be one of {True} or {False}.")

        lookup_leap = {True: 28,
                False: 29}

        if type(demand) is list:
            if len(demand)<12:
                raise Exception("Monthly demand list must have a length of 12.")
            
            print(f"Comparing list of monthly demands with supply timeseries. demand_timescale parameter is unused.")

            if demand_type == "daily_constant":
                print(f"Each monthly value in demand list will be applied as the daily demand for the respective month.")
                print(f"ignore_leap_years parameter is unused.")
                state = "monthly_constant_list"
            else:
                print(f"Each monthly total in demand list will be disaggregated to a daily demand for the respective month.")
                print(f"Leap years will be disaggregated assuming {lookup_leap[ignore_leap_years]} days in February.")
                state = "monthly_total_list"

        if type(demand) in [float, int]:
            print(f"Comparing provided demand with supply timeseries.")
            if (demand_timescale == "daily") or (demand_type == "daily_constant"):
                print(f"Assumed that daily demand is a constant {demand} ML/d.")
                if (demand_timescale == "daily") and (demand_type == "daily_constant"):
                    print(f"Only one of demand_timescale = {demand_timescale} and demand_type = {demand_type} were required.")
                    print(f"ignore_leap_years parameter is unused.")
                if (demand_timescale == "daily") and (demand_type != "daily_constant"):
                    print(f"demand_type and ignore_leap_years parameters are unused.")
                if (demand_type == "daily_constant") and (demand_timescale != "daily"):
                    print(f"demand_timescale and ignore_leap_years parameters are unused.")
                state="daily_constant"
            else:
                print(f"Assumed that {demand} ML is the total demand over the {demand_timescale} timescale. {demand} ML {demand_timescale} will be disaggregated to ML/d.")
                print(f"Leap years will be disaggregated assuming {lookup_leap[ignore_leap_years]} days in February.")
                if demand_timescale == "yearly":
                    state="yearly_total"
                if demand_timescale == "monthly":
                    state="monthly_total"

        self.demand=demand
        self.supply=supply
        self.demand_timescale=demand_timescale
        self.demand_type=demand_type
        self.ignore_leap_years=ignore_leap_years
        self.state=state
        

    def ReliabilityTS(self):
        """Returns demand as a timeseries for input to reliability statistics. Matches date range of supply timeseries input.

        Args:

        Returns:
            pd.Series: Demand timeseries for input to reliability stats
        """
        # If provided demand is a timeseries, just return timeseries.
        if self.state=="ts":
            common_dates=np.intersect1d(self.demand.index,self.supply.index)
            demand_ts=self.demand[common_dates]
            self.supply=self.supply[common_dates]
            return demand_ts

        else:
            # Copy date range of supply TS
            demand_ts=self.supply.copy(deep=True)

            #  Overwrite demand_ts with constant daily demand.
            if self.state=="daily_constant":
                demand_ts[:]=self.demand
                return demand_ts
            
            # Overwrite demand_ts with respective month constant daily demand
            if self.state=="monthly_constant_list":
                month_list=list(utils.get_month(demand_ts.index))
                demand_ts[:]=[self.demand[x-1] for x in month_list]
                return demand_ts

            # Overwrite demand_ts with respective total month demand disaggregated to daily.
            if self.state=="monthly_total_list":
                year_month=utils.get_year_and_month(demand_ts.index)
                if self.ignore_leap_years == False:
                    demand_ts[:]=[self.demand[int(x[5:7])-1]/calendar.monthrange(int(x[0:4]),int(x[5:7]))[1] for x in year_month] # If using leap years, divide by 28 or 29 Feb days depending on year
                else:
                    demand_ts[:]=[self.demand[int(x[5:7])-1]/calendar.monthrange(2002,int(x[5:7]))[1] for x in year_month] # If not using leap years, only divide by 28 Feb days
                return demand_ts
            
            # Overwrite demand_ts with total month demand disaggregated to daily.
            if self.state=="monthly_total":
                year_month=utils.get_year_and_month(demand_ts.index)
                if self.ignore_leap_years == False:
                    demand_ts[:]=[self.demand/calendar.monthrange(int(x[0:4]),int(x[5:7]))[1] for x in year_month] # If using leap years, divide by 28 or 29 Feb days depending on year
                else:
                    demand_ts[:]=[self.demand/calendar.monthrange(2002,int(x[5:7]))[1] for x in year_month] # If not using leap years, only divide by 28 Feb days
                return demand_ts

            # Overwrite demand_ts with total annual demand disaggregated to daily.
            if self.state=="yearly_total":
                if self.ignore_leap_years == False:
                    year_month=utils.get_year_and_month(demand_ts.index)
                    demand_ts[:]=[self.demand/(365+calendar.isleap(int(x[0:4]))) for x in year_month] # If using leap years, divide by 365 or 366 days depending on year
                else:
                    demand_ts[:]=self.demand/365 # If not using leap years, only ever divide by 365                    
                return demand_ts


    def MonthlyReliability(self, tol=1, allow_part_months=False):
        """Returns the monthly reliability statistic for a daily timeseries of demand and supply.

        Args:
            tol (float, optional): Percentage of demand treated as full demand. Defaults to 1 (100%).
            allow_part_months (bool, optional): Allow part months or only complete months. Defaults to False.

        Returns:
            float: Monthly reliability (%)
        """    
        # Pass input to reliability_ts first and enforce common date range.
        demand_ts=self.ReliabilityTS()

        # Collate timeseries data to monthly
        if allow_part_months:
            dem_month=demand_ts.groupby(utils.get_year_and_month(demand_ts.index)).sum()
            sup_month=self.supply.groupby(utils.get_year_and_month(self.supply.index)).sum()
        else:
            if self.supply.index[0][8:10]=="01": #0123-56-89
                #First date is the start of a month; use this as start date.
                start_date=self.supply.index[0]
            else:
                #Start on the first date of the next month
                start_date = utils.get_next_month_start(self.supply.index[0])
            if self.supply.index[-1] == utils.get_this_month_end(self.supply.index[-1]):
                end_date=self.supply.index[-1]
            else:
                end_date=utils.get_prev_month_end(self.supply.index[-1])
            demand_trim = demand_ts[start_date:end_date]
            supply_trim = self.supply[start_date:end_date]
            year_month = utils.get_year_and_month(demand_trim.index)
            dem_month=demand_trim.groupby(year_month).sum()
            sup_month=supply_trim.groupby(year_month).sum()

        # Check whether demand is met within given tolerance and express as a percentage
        rel=np.where((sup_month>=dem_month*tol),1,0).sum()/sup_month.count()
        return rel


    def AnnualReliability(self, tol=1, wy_month=7, allow_part_years=False):
        """Returns the annual reliability statistic for a daily timeseries of demand and supply.

        Args:       
            tol (float, optional): Percentage of demand treated as full demand. Defaults to 1 (100%).
            wy_month (int, optional): Water year start month. Defaults to 7.
            allow_part_years (bool, optional): Allow part water years or only complete water years. Defaults to False.

        Returns:
            float: Annual reliability (%)
        """
        
        # Pass input to reliability_ts first and enforce common date range.
        demand_ts=self.ReliabilityTS()
        
        # Collate timeseries data to annual
        if not allow_part_years:
            demand_ts = utils.crop_to_wy(demand_ts, wy_month)
            supply = utils.crop_to_wy(self.supply, wy_month)
        else:
            supply = self.supply
        if (len(demand_ts) == 0):
            return np.nan
        dem_annual=demand_ts.groupby(utils.get_wy(demand_ts.index, wy_month)).sum()
        sup_annual=supply.groupby(utils.get_wy(supply.index, wy_month)).sum()
        no_years=sup_annual.count()

        # Check whether demand is met within given tolerance and express as a percentage
        rel=np.where((sup_annual>=dem_annual*tol),1,0).sum()/no_years

        return rel



import os
import numpy as np
import pandas as pd
import subprocess
from odhpy import utils


class iqqm_out_reader:

    def __init__(self, iqqm_out_filepath) -> None:
        self.iqqm_out_filepath = iqqm_out_filepath
        self.iqqm_out_folder = os.path.dirname(self.iqqm_out_filepath)
        self.iqqm_out_basename = os.path.basename(self.iqqm_out_filepath)[:-4]
        self.required = {}
        self.available = {}
        self.__files_requiring_cleanup = []
        self.__search_available_data()


    def require(self, node:any=None, supertype:float=None, type:float=None, output:any=None):
        # Coherse node and output into string formats padded with zeros.
        node = None if node is None else f"000{node}"[-3:]
        output = None if output is None else f"00{output}"[-2:]
        # Now loop over all available records and identify the ones required by the user.
        for k, v in self.available.items():
            if ((node == None) or (node == v["node"])) and \
            ((supertype == None) or (supertype == v["supertype"])) and \
            ((type == None) or (type == v["type"])) and \
            ((output == None) or (output == v["output"])):
                self.required[k] = v


    def read(self, remove_temp_files=True, read_all_availabe=False) -> pd.DataFrame:
        if read_all_availabe:
            self.required = self.available
        self.__write_iqqmgui_lqn_file()
        self.__call_iqqmgui_lqn()
        answer = self.__read_csv()        
        if remove_temp_files: 
            self.__clean_up()
        return answer
    

    def __search_available_data(self):
        with open(self.iqqm_out_filepath, mode="r") as file:
            ss = file.readlines()
        ss2 = ss[2].split() #line 3 in the file
        # Read the recorder-flag matrix
        n_node_types = int(ss2[0])
        n_output_types = int(ss2[1])
        recorder_flags = []
        for i in range(n_node_types):
            temp = ss[3 + i].split()
            recorder_flags.append(temp[0:n_output_types])
        # Read the date range
        ssx = ss[n_node_types + 3].split()                          #01/01/1890 31/12/2008  0
        self.start_dt_str = ssx[0].replace('/',' ')
        self.end_dt_str = ssx[1].replace('/',' ')
        # Read all the nodes (loop over the nodes)
        for i in range(n_node_types + 4, len(ss)):
            temp = ss[i]
            if str.strip(temp) == "":
                break
            node_number = f"000{int(temp[0:3])}"[-3:]                #053
            node_name = str.strip(temp[3:20])                        #'Unallocated Irr'
            node_type = float(temp[20:])                             #8.3
            node_supertype = node_type.__floor__()                   #8
            for j in range(n_output_types):
                if recorder_flags[node_supertype][j] == "0":
                    continue
                recorder_number = f"000{j + 1}"[-2:]                 #03; recorder numbers start at 1
                temp = [node_number, recorder_number, node_name, node_type, node_supertype]                
                self.available[f"{node_number}_{recorder_number}.d"] = {
                    "node": node_number,
                    "supertype": node_supertype,
                    "type": node_type,
                    "output": recorder_number
                }


    def __write_iqqmgui_lqn_file(self) -> None:
        """Generates an iqqmgui lqn file so that we can use iqqm to extract data to csv."""
        self.lqn_filename = f"temp.run"
        self.lqn_filepath = f"{self.iqqm_out_folder}/{self.lqn_filename}"        
        with open(self.lqn_filepath, "w+", encoding='utf-8') as file:
            file.write("Listing file generated by odhpy\n")
            file.write(f"{self.start_dt_str} {self.end_dt_str} /\n")        #01 01 1890 31 12 2008 / start date, end date
            file.write(f"'{self.iqqm_out_basename}' /\n")                   #'O02l' / Name of IQN File
            file.write(f"{len(self.required)} 0 1 /\n")                     #17 0 1 /no files, no eqns, (no csv ?)
            i = 0
            for k, v in self.required.items():
                i += 1
                iqqm_ts_filepath = f"{self.iqqm_out_folder}/{k}"
                self.__files_requiring_cleanup.append(iqqm_ts_filepath)
                file.write(f"'{k}' 00 00 00 00 T / {[i]}\n")                #'O02l#030.01d' 00 00 00 00 T / [1]
                file.write(f"1 0 0 /\n")                                    #1 0 0 / 
                file.write(f"{v['node']} {v['output']} /\n")                #030 1 /
            self.csv_filename = f"temp.csv"
            self.csv_filepath = f"{self.iqqm_out_folder}/{self.csv_filename}"
            file.write(f"{self.csv_filename} /\n")                          #DW_Diversions.csv /
            file.write(f"1-{i} /\n")                                        #1-17 /
        self.__files_requiring_cleanup.append(self.lqn_filepath)


    def __call_iqqmgui_lqn(self) -> None:
        """Uses iqqmgui to extract data to csv."""
        process = subprocess.Popen(f"iqmgui {self.lqn_filename}", cwd=f"{self.iqqm_out_folder}")
        process.wait()
        self.__files_requiring_cleanup.append(self.csv_filepath)
        self.__files_requiring_cleanup.append(f"{self.iqqm_out_folder}/iqqmml.txt") # Artifact from running iqmgui


    def __read_csv(self) -> None:
        """Reads the csv into a dataframe."""
        df = pd.read_csv(self.csv_filepath)
        df.columns = ["Date"] + [c for c in df.columns[1:]]
        #df = utils.set_index_dt(df)
        df.set_index("Date", inplace=True)
        utils.assert_df_format_standards(df)
        return df
    

    def __clean_up(self) -> None:
        """Removes any file-artifacts created by this class."""
        for f in self.__files_requiring_cleanup:
            os.remove(f)
        self.__files_requiring_cleanup.clear()
    



import numpy as np
import pandas as pd


def derive_transformation_curves(original_ts: pd.Series, augmented_ts: pd.Series, season_start_months=[1,2,3,4,5,6,7,8,9,10,11,12], epsilon=1e-3) -> dict:
    """Returns a dictionary of exceedence-based transformation curves - one for each season 
    with the season's start month as the key. These are tables that map from exceedance 
    (cunnane plotting position as a fraction) to a scaling factor. These are intended to 
    be used to
    effectively summarise climate-change adjustments, and allow them to be transported from
    one timeseries to another.

    Args:
        original_ts (pd.Series): _description_
        augmented_ts (pd.Series): _description_
        season_start_months (list, optional): _description_. Defaults to [1,2,3,4,5,6,7,8,9,10,11,12].

    Returns:
        dict: _description_
    """
    df = pd.DataFrame()
    df["x"] = original_ts
    df["y"] = augmented_ts
    df = df.dropna() # Force common period
    answer = {}
    for i in range(len(season_start_months)):
        # Get a list of the months in this season
        start_month = season_start_months[i]
        season_len = (season_start_months + [m + 12 for m in season_start_months])[i + 1] - start_month
        months_in_this_season = [1,2,3,4,5,6,7,8,9,10,11,12,1,2,3,4,5,6,7,8,9,10,11,12][start_month - 1: start_month - 1 + season_len]
        # Find the data for this season
        df_m = df[[int(d[5:7]) in months_in_this_season for d in df.index]] #d[5:7] is the month part of the date string
        x = np.sort(df_m.x.values)
        y = np.sort(df_m.y.values)
        # The transformation factor is y/x except when the original value x is zero (<epsilon) in which case we default to 1.0
        f = np.where(x < epsilon, 1.0, y / x)
        n = len(x)
        ii = [i + 1 for i in range(n)] #index starting at 1
        p = [(i - 0.4)/(n + 0.2) for i in ii]
        answer[start_month] = [p,f]
    return answer     

    
def apply_transformation_curves(tranformation_curves: dict, series: pd.Series) -> pd.Series:
    """Applies seasonal transformation curves to an input series.
    Refer to the function 'derive_transformation_curves(...)'.

    Args:
        tranformation_curves (dict): _description_
        series (pd.Series): _description_

    Returns:
        pd.Series: _description_
    """
    dates = series.index
    answer = series.copy()    
    # Apply each transformation curves to the whole series. Splice the appropriate 
    # parts (seasons) into the 'answer' series as we go. 
    season_start_months = sorted(tranformation_curves.keys())
    for i in range(len(season_start_months)):
        # Identify the transform curve for this season
        start_month = season_start_months[i]
        t = tranformation_curves[start_month]
        xp = t[0] 
        fp = t[1]
        # Get a list of the months in this season
        season_len = (season_start_months + [m + 12 for m in season_start_months])[i + 1] - start_month
        months_in_this_season = [1,2,3,4,5,6,7,8,9,10,11,12,1,2,3,4,5,6,7,8,9,10,11,12][start_month - 1: start_month - 1 + season_len]
        # Find the data for this season
        m = len(series)
        season_dates = pd.Series([d for d in dates if int(d[5:7]) in months_in_this_season]) #d[5:7] is the month part of the date string
        values = answer[season_dates]
        # And get their ranks and plotting positions
        rank_starting_at_one = values.rank(ascending=True) # This function is nice because equal values are assigned the same (averaged) rank.
        n = len(values)
        p = [(r - 0.4)/(n + 0.2) for r in rank_starting_at_one] # plotting position
        f = np.interp(p, xp, fp) # interpolated scaling factors
        # Calcualte new values and update the answer
        new_values = pd.Series([values.iloc[i] * f[i] for i in range(n)], index=season_dates)
        answer.update(new_values)
    # Return a pd.Series so user can easily join it back into a dataframe
    return pd.Series(answer, index=dates, name=series.name)     
    
    
def derive_transformation_factors(original_ts: pd.Series, augmented_ts: pd.Series, season_start_months=[1,2,3,4,5,6,7,8,9,10,11,12], epsilon=1e-3) -> dict:
    """Returns a dictionary of transformation factors - one for each season 
    with the season's start month as the key. These scaling factors are intended to 
    be used to effectively summarise climate-change adjustments, and allow them to be 
    transported from one timeseries to another.

    Args:
        original_ts (pd.Series): _description_
        augmented_ts (pd.Series): _description_
        season_start_months (list, optional): _description_. Defaults to [1,2,3,4,5,6,7,8,9,10,11,12].
        epsilon: Threshold below which values are treated as zero, and the associated factor defaults to 1.

    Returns:
        dict: _description_
    """
    # Create a map of month -> season_start_month (for all months)
    month_to_season_map = {}
    key = max(season_start_months)
    for m in [1,2,3,4,5,6,7,8,9,10,11,12]:
        if m in season_start_months:
            key = m
        month_to_season_map[m] = key
    # Put the data in a dataframe, and groupby season start month
    df = pd.DataFrame()
    df["x"] = original_ts
    df["y"] = augmented_ts
    df = df.dropna() # Force common period
    df['m'] = df.index.month
    df['s'] = df['m'].map(month_to_season_map)
    df2 = df.groupby('s').agg('sum')
    df2['f'] = np.where(df2.x < epsilon, 1.0, df2.y / df2.x)
    return df2['f'].to_dict()

    
def apply_transformation_factors(tranformation_factors: dict, series: pd.Series) -> pd.Series:
    """Applies seasonal transformation factors to an input series.
    Refer to the function 'derive_transformation_factors(...)'.

    Args:
        tranformation_curves (dict): _description_
        series (pd.Series): _description_

    Returns:
        pd.Series: _description_
    """
    # Create a map of month -> factor (containing all months)
    season_start_months = sorted(tranformation_factors.keys())
    month_to_factor_map = {}
    key = max(season_start_months)
    for m in [1,2,3,4,5,6,7,8,9,10,11,12]:
        if m in season_start_months:
            key = m
        month_to_factor_map[m] = tranformation_factors[key]
    # Apply transformation factors to the whole series. Splice the appropriate 
    df = pd.DataFrame()
    df['x'] = series
    df['m'] = df.index.month
    df['f'] = df['m'].map(month_to_factor_map)
    df['y'] = df['x'] * df['f']
    answer = df['y']
    answer.name = series.name
    return answer

    
    
    


import pandas as pd
from odhpy import utils
from pandas.api.types import is_datetime64_any_dtype as is_datetime
from datetime import datetime


def assert_df_format_standards(df: pd.DataFrame):
    """
    Args:
        df (pd.DataFrame): _description_
    """
    violations = check_df_format_standards(df)
    if len(violations) > 0:
        raise Exception(f"Dataframe does not meet odhpy format standards.\n {violations[0]}")
    

def crop_to_wy(df: pd.DataFrame, wy_month=7):
        start_date=utils.get_wy_start_date(df, wy_month)
        end_date=utils.get_wy_end_date(df, wy_month)
        if isinstance(start_date, datetime): #handle datetime inputs
            start_date = start_date.strftime(r"%Y-%m-%d")
            end_date = end_date.strftime(r"%Y-%m-%d")
        return df.loc[(df.index >= start_date) & (df.index <= end_date)]
        

def check_df_format_standards(df: pd.DataFrame):
    """
    Checks if a given dataframe meets standards generally requried by 
    odhpy functions. These standards include:
    - Dataframe is not none
    - Dateframe is not empty
    - Dataframe index name is "Date"
    - Dataframe index values are daily sequential strings with the format "%Y-%m-%d"
    - Data columns all have datatype of double
    - Missing values are nan (not na, not -nan)

    Args:
        df (_type_): _description_

    Returns:
        bool: _description_
    """
    # - Dataframe is not none
    if df is None:
        return ["Dataframe is None"]
    # - Dataframe index name is "Date"
    if df.index.name != "Date":
        return ["Dataframe index name is not 'Date'"]
    # - Dataframe index values are daily sequential strings with the format "%Y-%m-%d"
    if len(df) > 0:
        # Check index values are strings
        if not isinstance(df.index[0], str):
            return [f"Index values are not strings: {type(df.index[0])}"]
        # Call function to determine string format
        str_format = utils.get_date_format(df.index[0])
        if str_format != r"%Y-%m-%d":
            return [f"Date string format should be '%Y-%m-%d' but is '{str_format}'"]
        # Check for sequential dates
        start_datetime = datetime.strptime(df.index[0], str_format)
        expected_date_strings = utils.get_dates(start_date=start_datetime, days=len(df), str_format=r"%Y-%m-%d")
        for i in range(len(df)):
            if df.index.values[i] != expected_date_strings[i]:
                return [f"Expected date string '{expected_date_strings[i]}' but found '{df.index.values[i]}' at index {i}"]
    # - Data columns all have datatype of double
    for c in df.columns:
        data_type = df[c].dtypes
        if (data_type != 'int64') and (df[c].dtypes != 'float64'):
            return [f"Column '{c}' is not int64 or float64: {data_type}"]
    # - Missing values are nan (not na, not -nan)
    return []


def set_index_dt(df: pd.DataFrame, dt_values=None, start_dt=None, **kwargs):
    """
    Returns a dataframe with datetime index. Useful for converting ODHPY dataframes to datetime
    as needed. WARNING: THE RETURNED DATAFRAME WILL BE INCONSISTENT WITH ODHPY STANDARD WHICH USES STRING DATES.
    
    If no optional arguments are provided, the function will look for a column named "date" (not 
    case-sensitive) within the input dataframe. Otherwise dt_values or start_dt (assumes daily)
    may be provided. 

    Args:
        df (pd.DataFrame): _description_
        dt_values (_type_, optional): _description_. Defaults to None.
        start_dt (_type_, optional): _description_. Defaults to None.
    """
    if (df is None) or (len(df.columns) == 0):
        raise Exception("The dataframe is None, or has no columns.")
    answer = None
    n_days = len(df)

    # If start_dt or dt_values was provided, then we go ahead and use those. This could 
    # potentially override an existing "Date" column, but I assume thats what the user 
    # wants if used in that way.
    if start_dt != None:
        df["Date"] = utils.get_dates(start_dt, days=n_days)
    elif dt_values != None:
        if len(dt_values) < n_days:
            raise Exception("dt_values is shorter than the dataframe.") 
        df["Date"] = dt_values[:n_days]

    # Now if the index is datetimes (whether it already was, or was created above) we just 
    # need to make sure it is named "Date" and return the df.
    if is_datetime(df.index):
        df.index.name = "Date"
        return df

    # If the index is named "Date" and is strings, we try to convert it to datetimes.
    if (df.index.name is not None) and (df.index.name.upper() == "DATE"):
        df.index.name = "Date"
        df.index = pd.to_datetime(df.index, **kwargs)
        return df

    # If there's no column named exactly "Date" (case sensitive) then we look for something 
    # ignoring whitespaces and capitalization and rename it exactly to "Date".
    if not "Date" in df.columns:
        for c in df.columns:
            if c.lower().strip() == "date":
                df = df.rename(columns={c: "Date"})
    
    # If there's still no "Date" column let's use the first column instead.
    if not "Date" in df.columns:
        df = df.rename(columns={df.columns[0]: "Date"})
    
    # Now, after all that, if there's a column called exactly "Date", we try to convert 
    # values to datetimes. 
    df["Date"] = pd.to_datetime(df["Date"], **kwargs)
    answer = df.set_index("Date")
    return answer


def datetimes_to_strings(v, str_format: str=r"%Y-%m-%d", **kwargs):
    """Converts a list of datetimes to strings using the given format.

    Args:
        v (_type_): _description_
        str_format (str, optional): _description_. Defaults to r"%Y-%m-%d".

    Returns:
        _type_: _description_
    """
    answer = [d.strftime(str_format) for d in v]
    return answer


def strings_to_datetimes(v, **kwargs):
    """Converts a list of strings to datetimes.

    Args:
        v (_type_): _description_

    Returns:
        _type_: _description_
    """
    answer = pd.to_datetime(v, **kwargs)
    return answer


def convert_index_to_datetime(df: pd.DataFrame, **kwargs):
    """
    Converts the index to pandas datetime. Accepts a dataframe with a index as datetime or strings.
    
    Args:
        df (pd.DataFrame): _description_ 
        **kwargs: _description_       
    """
    if (df is None) or (len(df.columns) == 0):
        raise Exception("The dataframe is None, or has no columns.")
    n_days = len(df)
    
    # Force the index name to be "Date"
    df.index.name = "Date"    
    
    if (n_days == 0) or (is_datetime(df.index)):
        # Nothing needed to be done
        pass
    elif (type(df.index[0]) is str):
        # Try to convert strings to datetimes.
        #df.index = pd.to_datetime(df.index, **kwargs)
        df.index = strings_to_datetimes(df.index, **kwargs)
    else:
        raise Exception("The index is not strings or datetimes.")
    return df


def convert_index_to_string(df: pd.DataFrame, str_format: str=r"%Y-%m-%d", **kwargs):
    """
    Converts the index to strings. Accepts a dataframe with a index as datetime or strings.
    
    Args:
        df (pd.DataFrame): _description_ 
        str_format (str, optional): _description_. Defaults to r"%Y-%m-%d".
        **kwargs: _description_       
    """    
    if (df is None) or (len(df.columns) == 0):
        raise Exception("The dataframe is None, or has no columns.")
    n_days = len(df)
    
    if (n_days == 0) or (type(df.index[0]) is str):
        # Nothing needed to be done
        pass
    elif is_datetime(df.index):
        # Try to convert datetimes to strings.
        new_index_values = [d.strftime(str_format) for d in df.index]
        df.index = new_index_values
    else:
        raise Exception("The index is not strings or datetimes.")

    # Force the index name to be "Date"
    df.index.name = "Date"    
    return df


def check_data_equivalence(df1, df2, check_col_order=True, threshold=1e-6, details:dict=None):
    """Checks if two numeric dataframes are the same. It checks the names & order of the columns, 
    the values of the index, and summary stats on all the data columns.

    Args:
        df1 (_type_): First dataframe.
        df2 (_type_): Second dataframe.
        check_col_order (bool, optional): Specifies whether column order is important or not. Defaults to True.
        threshold (float, optional): Numerical threshold for checking if stats are the same. Defaults to 1e-6.
        details (dict, optional): This is a dicitonary for returning detailed results to the user. Defaults to None.

    Returns:
        bool: _description_
    """
    if details is None:
        details = {}
    # Check if columns are the same
    c1 = df1.columns
    c2 = df2.columns
    differences = set(c1).difference(set(df2))
    if len(differences) > 0:
        temp = differences.intersection(set(c1))
        details[len(details)] = f"{len(temp)} columns are unique to df1: {temp}"
        temp = differences.intersection(set(c2))
        details[len(details)] = f"{len(temp)} columns are unique to df2: {temp}"
        return False
    # Check column order
    if check_col_order:
        for i in range(len(c1)):
            if c1[i] != c2[i]:
                details[len(details)] = f"Column order difference at col {i}: '{c1[i]}' != '{c2[i]}'"
                return False
    # Check row count
    if len(df1) != len(df2):
        details[len(details)] = f"Row count difference: '{len(df1)}' != '{len(df2)}'"
        return False
    # Check values of indices
    for i in range(len(df1)):
        if df1.index[i] != df2.index[i]:
            details[len(details)] = f"Indices not matching at row {i}: {df1.index[i]} != {df2.index[i]}"
            return False
    # Check the stats
    stats1 = df1.describe()
    stats2 = df2.describe()
    for c in c1:
        for stat_name in stats1.index:
            val1 = stats1.loc[stat_name][c]
            val2 = stats2.loc[stat_name][c]
            if abs(val1 - val2) > threshold:
                details[len(details)] = f"Column '{c}' has a different {stat_name}: {val1} != {val2}"
                return False            
    # All checks passed
    return True



import os
import numpy as np
import pandas as pd
from odhpy import utils



def read_res_csv(filename, custom_na_values=None, df=None, colprefix=None, allow_nonnumeric=False, **kwargs):
    """Reads a res csv data file into a DataFrame, and sets the index to the Date.

    Args:
        filename (_type_): _description_
        custom_na_values (_type_): A list of values to override the automatically-determined missing values. If None, the missing values will include any defined in the .res.csv file as well as ['', ' ', 'null', 'NULL', 'NAN', 'NaN', 'nan', 'NA', 'na', 'N/A' 'n/a', '#N/A', '#NA', '-NaN', '-nan'].

    Returns:
        _type_: _description_
    """
    # Handle custom na values
    if custom_na_values is None:
        na_values = ['', ' ', 'null', 'NULL', 'NAN', 'NaN', 'nan', 'NA', 'na', 'N/A' 'n/a', '#N/A', '#NA', '-NaN', '-nan']
    else:
        na_values = custom_na_values
    # If no df was supplied, instantiate a new one
    if df is None:
        df = pd.DataFrame()
    # Scrape through the header
    metadata_lines = []
    eoh_found = False
    with open(filename) as f:
        line = ""
        for line in f:
            metadata_lines.append(line)
            if line.strip().startswith("EOH"):
                eoh_found = True
                break
            if custom_na_values is None and line.strip().lower().startswith("missing data value,"):
                new_na_value = line.strip()[len("missing data value,"):] #e.g. "-9999"
                na_values.append(new_na_value)
    if not eoh_found:
        return None #maybe it's not a .res.csv
    col_header_line_number = len(metadata_lines) - 2
    lines_to_skip = [i for i in range(col_header_line_number)] + [col_header_line_number + 1]
    # Read the data
    temp = pd.read_csv(filename, na_values=na_values, skiprows=lines_to_skip)
    # Date index
    temp.set_index(temp.columns[0], inplace=True)
    temp.index = utils.standardize_datestring_format(temp.index)
    temp.index.name = "Date"
    # Check values
    if not allow_nonnumeric:
        for col in temp.columns:
            if not np.issubdtype(temp[col].dtype, np.number):
                raise Exception(f"ERROR: Column '{col}' is not numeric!")
    # Rename columns if required
    if colprefix is not None:
        for c in temp.columns:
            temp.rename(columns = {c:f"{colprefix}{c}"}, inplace = True)
    # Join to existing dataframe if required
    if df is None:
        df = temp
    else:
        if len(df) > 0:
            # Check that the dates overlap
            newdf_ends_before_df_starts = temp.index[0] < df.index[-1]
            df_ends_before_newdf_starts = df.index[-1] < temp.index[0]
            if newdf_ends_before_df_starts or df_ends_before_newdf_starts:
                raise Exception("ERROR: The dates in the new dataframe do not overlap with the existing dataframe!")
        df = df.join(temp, how="outer")
    # Return
    #utils.assert_df_format_standards(df)
    return df





# TODO: Write a function that generates node diagrams using dash.cytoscape (seems good, but you may find something better)
#
# https://dash.plotly.com/cytoscape/styling
# https://dash.plotly.com/cytoscape/layout

## Joel scribbles in "node_diagrams_testing.ipynb"



import os
import pandas as pd
import uuid
import shutil
import subprocess
from odhpy import utils
from .csv_io import *



def read_idx(filename, cleanup_tempfile=True) -> pd.DataFrame:
    """_summary_

    Args:
        filename (_type_): _description_
        cleanup_tempfile (bool, optional): _description_. Defaults to True.

    Returns:
        pd.DataFrame: _description_
    """
    if shutil.which('csvidx') is None:
        raise Exception("This method relies on the external program 'csvidx.exe'. Please ensure it is in your path.")
    # Extract data using csvidx
    temp_output_file = f"{uuid.uuid4().hex}.output.csv"
    command = f"csvidx {filename} {temp_output_file}"
    process = subprocess.Popen(command)
    process.wait()
    df = read_ts_csv(temp_output_file, skiprows=[1])
    # Cleanup
    if cleanup_tempfile:
        os.remove(temp_output_file)
        temp_log_file = temp_output_file + ".log"
        os.remove(temp_log_file)
    # Return
    return df



def write_idx(df, filename, cleanup_tempfile=True):
    """_summary_

    Args:
        df (_type_): _description_
        filename (_type_): _description_
    """
    if shutil.which('csvidx') is None:
        raise Exception("This method relies on the external program 'csvidx.exe'. Please ensure it is in your path.")
    temp_filename = f"{uuid.uuid4().hex}.tempfile.csv"
    write_area_ts_csv(df, temp_filename)
    command = f"csvidx {temp_filename} {filename}"
    process = subprocess.Popen(command)
    process.wait()
    if cleanup_tempfile:
        os.remove(temp_filename)



def write_area_ts_csv(df, filename, units = "(mm.d^-1)"):
    """_summary_

    Args:
        df (_type_): _description_
        filename (_type_): _description_
        units (str, optional): _description_. Defaults to "(mm.d^-1)".

    Raises:
        Exception: If shortenned field names are going to clash in output file.
    """
    # ensures dataframe adheres to standards
    utils.assert_df_format_standards(df)
    # convert field names to 12 chars and check for collisions
    fields = {}
    for c in df.columns:
        c12 = f"{c[:12]:<12}"
        if c12 in fields.keys():
            raise Exception(f"Field names clash when shortenned to 12 chars: {c} and {fields[c12]}")
        fields[c12] = c
    # create the header text
    header = f"{units}"
    for k in fields.keys():
        header += f',"{k}"'
    header += os.linesep
    header += "Catchment area (km^2)"
    for k in fields.keys():
        header += f", 1.00000000"
    header += os.linesep
    # open a file and write the header and the csv body
    with open(filename, "w+", newline='', encoding='utf-8') as file:        
        file.write(header)
        df.to_csv(file, header=False, na_rep=' NaN')
        
        


import matplotlib.pyplot as plt
import plotly.express as px
import pandas as pd



def plot_flowx(df, labels=None):
    """_summary_
    Ref: https://www.youtube.com/watch?v=6AurbMHGqBY

    Args:
        df (_type_): _description_
        labels (_type_, optional): _description_. Defaults to None.

    Returns:
        _type_: _description_
    """
    fig = px.line(df)
    fig.update_layout(
        yaxis_title_text = "Flow [ML/d]",
        yaxis_type = "log",
    )
    return fig


def plot_flow(df, labels=None):
    #plt.rcParams['figure.figsize'] = [12, 8]
    fig, ax = plt.subplots()
    for i in range(len(df.columns)):
        col = df.columns[i]
        if labels is not None:
            lab = labels[i]
        else:
            lab = col
        ax.plot(df[col], label=lab)
    ax.legend()
    ax.set_ylabel("Flow [ML/d]")
    ax.set_yscale('log')
    ax.grid(True)
    ax.set_ylim(1, None)
    #date_range = [datetime.date(2014, 1, 1), datetime.date(2022, 1, 1)]
    #ax.set_xlim(date_range)
    return fig, ax


def plot_exceedence(df, labels=None):
    fig, ax = plt.subplots()
    df_exceedence = df.dropna()
    nn = len(df_exceedence)
    index_starting_at_one = [i + 1 for i in range(nn)]
    df_exceedence["Exceedence"] = [100 * (r - 0.4)/(nn + 0.2) for r in index_starting_at_one]
    df_exceedence.set_index("Exceedence", inplace=True)
    for i in range(len(df_exceedence.columns)):
        col = df_exceedence.columns[i]
        if labels is not None:
            lab = labels[i]
        else:
            lab = col
        df_exceedence[col] = df_exceedence[col].sort_values(ascending=False).values
        ax.plot(df_exceedence[col], label=lab)
    ax.legend()
    ax.set_ylabel("Flow [ML/d]")
    ax.set_xlabel("Exceedence probability [%]")
    ax.set_yscale('log')
    ax.grid(True)
    ax.grid(True, which='minor')
    ax.set_ylim(1, None)
    ax.set_xlim(-2,102)
    return fig, ax



import pandas as pd
import altair as alt
import numpy as np
from typing import Union
from odhpy import utils
from odhpy import io
from odhpy import trans
import calendar
from datetime import datetime, timedelta

def wy_event_heatmap(df=Union[pd.DataFrame,pd.Series],criteria=1,y_title="Series",pass_label="Pass",fail_label="Fail",pass_colour="white",fail_colour="red",width=None,height=None,stroke='black',strokeWidth=0.3):
    """Returns an Altair heatmap chart from a timeseries input of event count by Water Year.

    Args:
        df (DataFrame | Series): Annual (WY) index timeseries of no. of 'events' in a year i.e. output of StorageLevelAssessment.AnnualDaysBelowSummary()
        criteria (int, optional): Minimum integer in df that triggers a 'failure'. Defaults to 1.
        y_title (str, optional): Y axis title. Defaults to "Series".
        pass_label (str, optional): Label for years that don't meet the criteria. Defaults to "Pass".
        fail_label (str, optional): Label for years that meet the criteria. Defaults to "Fail".
        pass_colour (str, optional): Colour for years that don't meet the criteria. Defaults to "white".
        fail_colour (str, optional): Colour for years that meet the criteria. Defaults to "red".
        width (int, optional): Width of chart output. Defaults to None.
        height (int, optional): Height of chart output. Defaults to None.
        stroke (str, optional): Colour of rectangle outline. Defaults to black.
        strokeWidth (float, optional): Width of rectangle outline. Defaults to 0.3.

    Returns:
        altair.Chart: Returns an Altair chart object
    """

    # If df input is a Series, first convert to DataFrame
    if type(df) is pd.Series:
        df=pd.DataFrame(df)

    # df column names must be str for Altair
    df.columns=[str(x) for x in df.columns]
    columns=df.columns.values


    # Lookup table to convert count of events by WY to 'pass' or 'fail
    lookup = pd.DataFrame({
    'key': ['1', '0'],
    'y': [fail_label,pass_label]
    })

    domain = [pass_label, fail_label]
    range = [pass_colour, fail_colour]

    # Altair needs vega expression. Generate based on criteria input
    vega_expr_str = 'if(datum.value>=' + str(criteria) + ',1,0)'

    # Default width and height expression
    if width == None:
        width = df.shape[0]*10
    if height == None:
        height = len(columns)*25
    
    chart_wy_event = alt.Chart(df.reset_index() # Reset index so that Altair can access Date index
    ).transform_fold( # Fold individual columns into single column
        columns,
        as_=['column','value']
    ).transform_calculate( # Convert count of WY events to 1 or 0 based on criteria
        value=vega_expr_str
    ).transform_lookup( # Convert 1 and 0 to pass and fail
        lookup='value',
        from_=alt.LookupData(lookup, key='key', fields=['y'])
    ).mark_rect(stroke=stroke, strokeWidth=strokeWidth
    ).encode(
        alt.X('index:O', title='Water Year'),
        alt.Y('column:N',sort=None, title=y_title),
        color=alt.Color('y:N', scale=alt.Scale(domain=domain,range=range), title="Key")
    ).properties(
        width=width,
        height=height
    )

    return chart_wy_event

def pyblo(dflist: list, sites: list, series: list, wy_month=1, site_order=None, series_order=None, colours=None, start_date=None, end_date=None, series_label="Series", site_label=None, label_freq=10, subtitle=True, width=None, height=35, stroke_width=0.5, stroke_colour='black',font_size=12, grid_colour='lightgrey'):
    """Returns a chart depicting available data by Water Year. Data is arranged by site with colour assigned by series. 

    Args:
        dflist (list): List of pd.Series data.
        sites (list): List of strings by which corresponding data in dflist is grouped by chart row.
        series (list): List of strings by which corresponding data in dflist is grouped by colour.
        wy_month (int, optional): Water year start month. Defaults to 1.
        site_order (list, optional): Optional order to arrange Sites by. Defaults to order in sites.
        series_order (list, optional): Optional order to arrange Series by. Defaults to order in series.
        colours (list, optional): Optional list of colours (arranged by series_order) to apply to Series. Defaults to selection from "muted rainbow" colour scheme.
        start_date (DateTime, optional): Optional minimum date (%d/%m/%Y) to display. Defaults to None.
        end_date (DateTime, optional): Optional maximum date (%d/%m/%Y) to display. Defaults to None.
        series_label (str, optional): Optional Series label for legend. Defaults to "Series".
        site_label (str, optional): Optional Site y-axis title. Defaults to None.
        label_freq (int, optional): Optional frequency to display year label. Defaults to 10 (years).
        subtitle (bool, optional): Whether to show subtitle. Defaults to True.
        width (int, optional): Optional chart width parameter. Defaults to function of total years.
        height (int, optional): Optional chart height parameter. Represents height per facet/site. Defaults to 35.
        stroke_width (float, optional): Optional width of bar outline. Defaults to 0.5.
        stroke_colour (str, optional): Optional colour of bar outline. Defaults to 'black'.
        font_size (int, optional): Optional font size to apply to all text. Defaults to 12.
        grid_colour (str, optional): Optional colour of gridlines. Defaults to 'lightgrey'.

    Returns:
        altair.Chart: Returns an Altair chart object
    """

    # Check that inputs are the same length
    if len(set([len(dflist),len(sites),len(series)])) != 1:
        raise Exception("All input lists must be of the same length")
    
    # Check that dflist is a list of pd.Series
    if (set([type(x) for x in dflist]) != {pd.Series}) or (set([utils.get_date_format(x.index[0]) for x in dflist]) != {r'%Y-%m-%d'}):
        raise Exception("All dflist entries must be a single column of a datetime-indexed dataframe (pd.Series)") 

    ## Initialise inputs
    using_end_year=False ## README: Assume False for Chas' get_wy commit
    date_fmt=r'%Y-%m-%d'

    ## Month lookup
    months = ["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]  

    # Get list of sites from first element of input dict   
    sitelist_unique = pd.Series(sites).unique()   # list of unique sites to plot

    # Get list of series from second element of input dict
    serieslist_unique = pd.Series(series).unique()   # list of unique series to plot

    # If colours array not provided, assign default colour scheme
    if colours == None:
        colours = ["#72AC57", "#639188", "#5C77B0", "#765EA4", "#98467A", "#BA3438", "#CF512B", "#E18433"]  # muted rainbow
    
    # If order array not provided, assign default order
    if series_order == None:
        series_order = serieslist_unique

    if site_order == None:
        site_order = sitelist_unique

    # Generate Vega expression for x-axis label frequency
    label_expr="year(datum.value) % " + str(label_freq) + " ? null : datum.label"

    # Rename data series columns for Altair folding (can't handle spaces/duplicates)
    [x.rename('Series'+str(i),inplace=True) for i,x in enumerate(dflist)]
    columnlist=[x.name for x in dflist]

    # Apply replaced series names to 'series_order' array
    index=[]
    for i,x in enumerate(series_order):
        for j,y in enumerate(series):
            if (x==y) and (j not in index):
                index.append(j)
                
    for i,x in enumerate(columnlist):
        if i not in index:
            index.append(i)

    series_order_ren = [columnlist[i] for i in index]

    # Get start and end dates of common index
    start=min([dflist[i].index[0] for i,x in enumerate(dflist)])
    end=max([dflist[i].index[-1] for i,x in enumerate(dflist)])

    # Create date range over combined indexes
    dates=utils.get_dates(start_date=datetime.strptime(start, date_fmt),end_date=datetime.strptime(end, date_fmt)+timedelta(1),str_format=date_fmt)

    # Combine all data into single df
    df_chart=pd.DataFrame(index=dates)
    df_chart=pd.concat([df_chart]+dflist,axis=1)

    # Replace actual data with respective df column source
    for i in columnlist:
        df_chart[i]=[(x) if (np.isnan(x)) else (i) for x in df_chart[i]]    

    df_chart_stack=pd.DataFrame(index=df_chart.index)

    for i in site_order:
        # Find series that are relevant to site site_order[i]
        relevant = [columnlist[j] for j,x in enumerate(sites) if x==i]

        # Find the relevant series that appears first in series_order list
        first_series = min([series_order_ren.index(x) for x in relevant])
        
        # Set the base series df for each site to begin filling from other series
        df_chart_stack[i]=df_chart[series_order_ren[first_series]]
        
        for x in series_order_ren:
            if x in relevant:
                # Replace NaNs with data from other columns where available, order determined by series_order.
                df_chart_stack[i]=df_chart_stack[i].fillna(df_chart[x])

    df_chart_stack_annual_list=[]

    for i in site_order:
        # Get WY count of series values present in each site df
        temp = df_chart_stack.groupby([utils.get_wy(df_chart_stack.index,wy_month), i]).size()
        temp = temp.reset_index().rename(columns={'level_0': 'index', i: 'Series', 0: 'Count'})

        # Calculate number of days in the WY
        if using_end_year:
            if wy_month>2:
                temp["Day_count"]=[366 if calendar.isleap(x) else 365 for x in temp["index"]]
            else:
                temp["Day_count"]=[366 if calendar.isleap(x-1) else 365 for x in temp["index"]]
        else:
            if wy_month>2:
                temp["Day_count"]=[366 if calendar.isleap(x+1) else 365 for x in temp["index"]]
            else:
                temp["Day_count"]=[366 if calendar.isleap(x) else 365 for x in temp["index"]]        

        # Convert count of data to percentage of year
        temp["Perc"]=temp["Count"].divide(temp["Day_count"].values, axis=0)
        temp["Site"]=i
        df_chart_stack_annual_list.append(temp)

    # Temporary fix: pylance is showing pd.concat unreachable
    def my_concat(list):
        cat=pd.concat(list)
        return cat

    # Combine site DFs into single 'long-format' df for Altair
    df_chart_stack_annual=my_concat(df_chart_stack_annual_list)

    # Convert index to pd.Timestamp for Altair
    df_chart_stack_annual["index"] = [pd.Timestamp(x,1,1) for x in df_chart_stack_annual["index"].to_list()]    

    # Lookup table to assign series from data input name (df column)
    lookup_colour = pd.DataFrame({
        'key': columnlist,
        'c': series
    })

    # Default domain expression
    end_date_year=df_chart_stack_annual["index"].max().year

    if start_date == None:
        start_date = df_chart_stack_annual["index"].min()
    else:
        start_date=datetime.strptime(start_date, date_fmt)
        start_date_wy = (start_date.year - 1) if (start_date.month < wy_month) else (start_date.year)                
        start_date = pd.Timestamp(start_date_wy,1,1)
    if end_date == None:
        end_date = df_chart_stack_annual["index"].max().replace(year=end_date_year+1)
    else:
        end_date=datetime.strptime(end_date, date_fmt)
        end_date_wy = (end_date.year - 1) if (end_date.month < wy_month) else (end_date.year)
        end_date = pd.Timestamp(end_date_wy+1,1,1)

    # Default width and height expression
    no_years = end_date.year-start_date.year+1
    if width == None:
        width = no_years*12       

    # Show subtitle if parameter is True
    if subtitle:    
        subtitle = ["Annual (" + months[wy_month-1] + " to " + months[wy_month-2] + ") percentage of available data: " + format(start_date.replace(month=wy_month),"%d/%m/%Y") + " - " + format(end_date.replace(month=wy_month)-timedelta(1),"%d/%m/%Y")]
    else:
        subtitle=[]

    chart=alt.Chart(df_chart_stack_annual
        ).transform_lookup( # Lookup data column series
            lookup='Series',
            from_=alt.LookupData(lookup_colour, key='key', fields=['c'])
        ).mark_bar(
            width=alt.RelativeBandSize(1),
            clip=True
        ).encode(
            x=alt.X(
                'index:T',
                timeUnit='year',
                axis=alt.Axis(
                    ticks=False,
                    tickCount=999999,
                    orient='top',
                    labelAngle=90,
                    title=None,
                    grid=True,
                    gridColor=grid_colour,
                    labelPadding=5,
                    labelExpr=label_expr),
                scale=alt.Scale(domain=[start_date,end_date])),
            y=alt.Y(
                'Perc:Q',
                title=None,
                axis=alt.Axis(
                    format=".1%",
                    tickCount=0)).stack('zero').scale(domain=[0, 1]),
            strokeWidth=alt.condition('datum.Perc > 0',alt.value(stroke_width),alt.value(0)), # don't show tiny bars for years with no data
            stroke=alt.value(stroke_colour),
            color=alt.Color('c:N',
                sort=series_order,
                title=series_label,
                scale=alt.Scale(range=colours)),
            order=alt.Order('color_c_sort_index:Q'),
            tooltip=[
                alt.Tooltip(
                    'Perc:Q',
                    format='.1%',
                    title="Percentage"),
                    alt.Tooltip(
                        'c:N',
                        title=series_label),
                    alt.Tooltip(
                        'index:T',
                        title="Water Year",
                        format='%Y')],
            row=alt.Row(
                field='Site',
                sort=site_order,
                spacing=0,
                title=site_label,
                header=alt.Header(
                    labelAngle=0,
                    labelAlign='left',
                    labelOrient='left'))
        ).properties(
            height=height,
            width=width,
            title=alt.TitleParams(
                subtitle,
                baseline='bottom',
                orient='bottom',
                anchor='start',
                fontWeight='normal',
                fontStyle='italic',
                fontSize=font_size*0.9,
            offset=font_size*0.9*2)
        ).configure_view(
            stroke='black'
        ).configure_axis(
            labelFontSize=font_size,
            titleFontSize=font_size
        ).configure_headerRow(
            labelFontSize=font_size,
            titleFontSize=font_size 
        ).configure_legend(
            labelFontSize=font_size,
            titleFontSize=font_size 
        )
    
    return chart

def exceedence_plot(df:pd.DataFrame,yLabel='Flow (ML/d)',xLog=False,yLog=False,legendTitle='Data set',plotWidth=500,plotHeight=300,plottingPosition="cunnane"):
    """
    Exceedance plot of timeseries data.

    Args:
        df (pd.DataFrame): Dataframe of daily timeseries data.
        yLabel (str, optional): Y axis title. Defaults to 'Flow (ML/d)'.
        xLog (bool, optional): X axis log flag. Defaults to False.
        yLog (bool, optional): Y axis log flag. Defaults to False.
        legendTitle (str, optional): Legend title. Defaults to 'Data set'.
        plotWidth (int, optional): Chart width. Defaults to 500.
        plotHeight (int, optional): Chart height. Defaults to 300.
        plottingPosition (str, optional): Defaults to "cunnane". Other supported values: "weibull", "gringorten". See https://glossary.ametsoc.org/wiki/Plotting_position

    Returns:
        altair.Chart: Returns an Altair chart object
    """

    df_exceedence = df.dropna()
    nn = len(df_exceedence)
    df_exceedence["Exceedence"] = trans.get_exceedence_plotting_position(nn,plotting_position=plottingPosition)
    df_exceedence.set_index("Exceedence", inplace=True)

    for i in range(len(df_exceedence.columns)):
        col = df_exceedence.columns[i]
        lab = col
        df_exceedence[col] = df_exceedence[col].sort_values(ascending=False).values
    
    if xLog:
        xType="log"
    else:
        xType="linear"

    if yLog:
        yType="log"
    else:
        yType="linear"

    df_exc_melted = df_exceedence.reset_index().melt('Exceedence')
    df_exc_melted.rename(columns={'value':yLabel,'variable':legendTitle}, inplace=True)
    excPlot=alt.Chart(df_exc_melted).mark_line().encode(
        alt.X('Exceedence:Q',
          axis=alt.Axis(title='Exceedance probability (%)')).scale(type=xType),
        alt.Y(yLabel + ':Q',
          axis=alt.Axis(title=yLabel)).scale(type=yType),
        color=legendTitle,
        tooltip=[alt.Tooltip('Exceedence',format=',.1f',title='Exceedence (%)'),alt.Tooltip(yLabel,format=',.0f')]
    ).properties(width=plotWidth,height=plotHeight).interactive(bind_y=False)
    
    return excPlot


def daily_plot(df:pd.DataFrame,yLabel='Flow (ML/d)',legendTitle='Data set',plotWidth=500,plotHeight=300):
    """
    Daily plot of timeseries data.

    Args:
        df (DataFrame): Dataframe of daily timeseries data.
        yLabel (str, optional): Y axis title. Defaults to 'Flow (ML/d)'.
        legendTitle (str, optional): Legend title. Defaults to 'Data set'.
        plotWidth (int, optional): Chart width. Defaults to 500.
        plotHeight (int, optional): Chart height. Defaults to 300.

    Returns:
        altair.Chart: Returns an Altair chart object
    """

    alt.data_transformers.disable_max_rows()

    df_melted = df.reset_index().melt('Date')
    df_melted.rename(columns={'value':yLabel,'variable':legendTitle}, inplace=True)
    dailyPlot=alt.Chart(df_melted).mark_line().encode(
        alt.X('Date:T',
          axis=alt.Axis(title='Date')),
        alt.Y(yLabel + ':Q',
          axis=alt.Axis(title=yLabel)),
        color=legendTitle,
        tooltip=[alt.Tooltip(legendTitle),alt.Tooltip('Date',title='Date'),alt.Tooltip(yLabel,format=',.0f')]
    ).properties(width=plotWidth,height=plotHeight).interactive(bind_y=False)
    
    return dailyPlot


def annual_plot(df:pd.DataFrame,yLabel='Flow (ML/a)',legendTitle='Data set',wyStartMonth=7,plotWidth=500,plotHeight=300):
    """
    Annual plot of timeseries data.

    Args:
        df (DataFrame): Dataframe of daily timeseries data.
        yLabel (str, optional): Y axis title. Defaults to 'Flow (ML/a)'.
        legendTitle (str, optional): Legend title. Defaults to 'Data set'.
        wyStartMonth (int, optional): Water year start month. Defaults to 7.
        plotWidth (int, optional): Chart width. Defaults to 500.
        plotHeight (int, optional): Chart height. Defaults to 300.

    Returns:
        altair.Chart: Returns an Altair chart object
    """

    alt.data_transformers.disable_max_rows()
    xLabel='Water year'

    if wyStartMonth==1: xLabel='Water year (Jan to Dec)'
    if wyStartMonth==2: xLabel='Water year (Feb to Jan)'
    if wyStartMonth==3: xLabel='Water year (Mar to Feb)'
    if wyStartMonth==4: xLabel='Water year (Apr to Mar)'
    if wyStartMonth==5: xLabel='Water year (May to Apr)'
    if wyStartMonth==6: xLabel='Water year (Jun to May)'
    if wyStartMonth==7: xLabel='Water year (Jul to Jun)'
    if wyStartMonth==8: xLabel='Water year (Aug to Jul)'
    if wyStartMonth==9: xLabel='Water year (Sep to Aug)'
    if wyStartMonth==10: xLabel='Water year (Oct to Sep)'
    if wyStartMonth==11: xLabel='Water year (Nov to Oct)'
    if wyStartMonth==12: xLabel='Water year (Dec to Nov)'

    start_date=utils.get_wy_start_date(df,wyStartMonth)
    end_date=utils.get_wy_end_date(df,wyStartMonth) + timedelta(days=1) #Add day because last index in .loc is not included
    cropped_df=df.loc[utils.get_dates(start_date,end_date)]
    annual_cropped_df=cropped_df.groupby(utils.get_wy(cropped_df.index, wyStartMonth)).sum(min_count=365)

    df_melted = annual_cropped_df.reset_index().melt('index')
    df_melted.rename(columns={'index':'Water year','value':yLabel,'variable':legendTitle}, inplace=True)
    annualPlot=alt.Chart(df_melted).mark_line().encode(
        alt.X('Water year',
          axis=alt.Axis(title=xLabel,format='.0f')),
        alt.Y(yLabel + ':Q',
          axis=alt.Axis(title=yLabel)),
        color=legendTitle,
        tooltip=[alt.Tooltip(legendTitle),alt.Tooltip('Water year'),alt.Tooltip(yLabel,format=',.0f')]
    ).properties(width=plotWidth,height=plotHeight).interactive(bind_y=False)
    
    return annualPlot

def residual_mass_curve(df:pd.DataFrame,yLabel='Flow residual mass (ML)',legendTitle='Data set',plotWidth=500,plotHeight=300):
    """
    Residual mass curve of timeseries data.

    Args:
        df (DataFrame): Dataframe of daily timeseries data.
        yLabel (str, optional): Y axis title. Defaults to 'Flow residual mass (ML)'.
        legendTitle (str, optional): Legend title. Defaults to 'Data set'.
        plotWidth (int, optional): Chart width. Defaults to 500.
        plotHeight (int, optional): Chart height. Defaults to 300.

    Returns:
        altair.Chart: Returns an Altair chart object
    """

    alt.data_transformers.disable_max_rows()

    residualMass=df.copy()
    
    for i in range (0,len(df.columns)):
        col=df.columns[i]
        avg=df[col].mean()
        cumDiff=0
        tempResidMass=[]
        for j in range (0,len(residualMass)):
            if(pd.isna(residualMass[col][j])):
                tempResidMass.append(float("nan"))
            else:
                todayDiff=residualMass[col][j]-avg
                cumDiff+=todayDiff
                tempResidMass.append(cumDiff)
        residualMass[col + ' residual mass']=tempResidMass
        residualMass.drop(columns=[col],inplace=True)

    df_melted = residualMass.reset_index().melt('Date')
    df_melted.rename(columns={'value':yLabel,'variable':legendTitle}, inplace=True)
    residMassPlot=alt.Chart(df_melted).mark_line().encode(
        alt.X('Date:T',
          axis=alt.Axis(title='Date')),
        alt.Y(yLabel + ':Q',
          axis=alt.Axis(title=yLabel)),
        color=legendTitle,
        tooltip=[alt.Tooltip(legendTitle),alt.Tooltip('Date',title='Date'),alt.Tooltip(yLabel,format=',.0f')]
    ).properties(width=plotWidth,height=plotHeight).interactive(bind_y=False)
    
    return residMassPlot


import os
import numpy as np
import pandas as pd
from odhpy import utils
na_values = ['', ' ', 'null', 'NULL', 'NAN', 'NaN', 'nan', 'NA', 'na', 'N/A' 'n/a', '#N/A', '#NA', '-NaN', '-nan']



def read_iqqm_lqn_output(filename, col_name=None, df=None) -> pd.DataFrame:
    """Reads the output of IQQM listquan. This is format is the same for flows and diversions.

    Args:
        filename (_type_): _description_
        df (_type_, optional): _description_. Defaults to None.

    Returns:
        pd.DataFrame: _description_
    """
    # If no df was supplied, instantiate a new one
    if df is None:
        df = pd.DataFrame()
    # If no column name was specified, we use the base name of the file
    if col_name is None:
        col_name = os.path.basename(filename)
    # Read the data
    data_start_row=7
    temp = pd.read_csv(filename, skiprows=(data_start_row-2), sep=r'\s+', names=["Date", col_name], header=None)
    #temp = utils.set_index_dt(temp, format='%d/%m/%Y')
    temp.set_index(temp.columns[0], inplace=True)
    temp.index = utils.standardize_datestring_format(temp.index)
    temp.index.name = "Date"
    temp = temp.replace(r'^\s*$', np.nan, regex=True)
    df = df.join(temp, how="outer").sort_index()
    # TODO: THERE IS NO GUARANTEE THAT THE DATES OVERLAP, THEREFORE WE MAY END UP WITH A DATAFRAME WITH INCOMPLETE DATES
    # TODO: I SHOULD MAKE DEFAULT BEHAVIOUR AUTO-DETECT FORMAT DEPENDING ON *TYPE* AND *LOCATION* OF DELIMIT CHARS
    # TODO: In the meantime we use the below to assert that the format of the resulting df meets our minimum standards.
    utils.assert_df_format_standards(df)
    return df
    
    








from datetime import datetime, timedelta
import pandas as pd
import numpy as np




def get_date_format(date_str):
    """
    Trial and error approach to determining the date format of a date string.
    """
    for date_fmt in [r'%Y-%m-%d', r'%d/%m/%Y', r'%d/%m/%Y %H:%M', r'%d/%m/%Y %H:%M:%s']:
        try:
            dt = datetime.strptime(date_str, date_fmt)
            return date_fmt
        except ValueError:
            pass
    raise ValueError('Invalid date format for "{}"'.format(date_str))


def standardize_datestring_format(values):
    """
    Converts a list of date strings into a list of date strings in the format YYYY-MM-DD.
    Tested over the range 0001-01-01 to 9999-12-31.
    """
    date_fmt = get_date_format(values[0])
    np_dates = to_np_datetimes64d(values, date_fmt=date_fmt)
    # start_date = datetime.strptime(values[0], date_fmt) #dateutil.parser.parse(values[0], format=fmt)
    # end_date = datetime.strptime(values[-1], date_fmt) + timedelta(days=1)
    # np_datetimes64s = np.arange(start_date, end_date, dtype='datetime64[D]')
    # if (len(np_datetimes64s) != len(values)):
    #     raise Exception(f"ERROR: Expected {len(np_datetimes64s)} dates between {start_date} and {end_date} but found {len(values)}.")
    answer = [str(t) for t in np_dates]
    return answer


def to_np_datetimes64d(values, date_fmt=r'%Y-%m-%d'):
    start_date = datetime.strptime(values[0], date_fmt)
    end_date = datetime.strptime(values[-1], date_fmt) + timedelta(days=1)
    np_dates = np.arange(start_date, end_date, dtype='datetime64[D]') #Assumes the dates are consecutive!
    if (len(np_dates) != len(values)):
        raise Exception(f"ERROR: Expected {len(np_dates)} dates between {start_date} and {end_date} but found {len(values)}.")
    return np_dates


def get_wy(dates, wy_month=7, using_end_year=False):
    """
    Returns water years, as a list of ints, for a given array of dates. Use this to
    add water year info into a pandas DataFrame. 

    The default (using_end_year=False) aligns water years with the primary water 
    allocation at the start of the water year. The alternative (using_end_year==True) 
    follows the convention used for fiscal years whereby water years are labelled 
    based on their end dates. Using the fiscal convention, the 2022 water year is 
    from 2021-07-01 to 2022-06-30 inclusive.
    """
    np_dates = to_np_datetimes64d(dates)
    #d.astype('datetime64[Y]').astype(int) + 1970     #<---- this gives the year
    #d.astype('datetime64[M]').astype(int) % 12 + 1   #<---- this gives the month
    # TODO: the below implementation was originally written for pd.Timestamp, not np.datetime64d. It may be possible to simplify it.
    if using_end_year:
        answer = [(d.astype('datetime64[Y]').astype(int) + 1970) if (d.astype('datetime64[M]').astype(int) % 12 + 1) < wy_month else (d.astype('datetime64[Y]').astype(int) + 1970) + 1 for d in np_dates]
    else:
        answer = [(d.astype('datetime64[Y]').astype(int) + 1970) - 1 if (d.astype('datetime64[M]').astype(int) % 12 + 1) < wy_month else (d.astype('datetime64[Y]').astype(int) + 1970) for d in np_dates]        
    return answer


def get_prev_month_end(stringdate):
    year_str = stringdate[:4] #"2021"
    month_str = stringdate[5:7] #"04"    
    day_str = "31" #default which covers most months
    #
    #Go to previous month
    month_int = (int(month_str) - 1)
    if (month_int == 0):
        month_str = "12"
        year_str = f"{(int(year_str) - 1):04d}"
    else:
        month_str = f"{month_int:02d}"    
    #
    #Set the day
    if month_str in ["04", "06", "09", "11"]:        
        day_str = "30"
    elif month_str == "02":
        year = int(year_str)        
        if year % 4 != 0:
            day_str = "28"
        elif year % 100 != 0:
            day_str = "29"
        elif year % 400 != 0:
            day_str = "28"
        else:
            day_str = "29"
    #
    #Return
    return f"{year_str}-{month_str}-{day_str}"


def get_this_month_end(stringdate):
    year_str = stringdate[:4] #"2021"
    month_str = stringdate[5:7] #"04"    
    day_str = "31" #default which covers most months
    if month_str in ["04", "06", "09", "11"]:        
        day_str = "30"
    elif month_str == "02":
        year = int(year_str)
        if year % 4 != 0:
            day_str = "28"
        elif year % 100 != 0:
            day_str = "29"
        elif year % 400 != 0:
            day_str = "28"
        else:
            day_str = "29"
    return f"{year_str}-{month_str}-{day_str}"


def get_next_month_start(stringdate):
    year_str = stringdate[:4] #"2021"
    month_str = stringdate[5:7] #"04"
    day_str = "01"
    if month_str == "12":
        year_str = f"{(int(year_str) + 1):04d}"
        month_str = "01"
    else:
        month_str = f"{(int(month_str) + 1):02d}"
    return f"{year_str}-{month_str}-{day_str}"


def get_year_and_month(v):
    """
    Returns year and month string e.g. "2022_01", as a list for a given array of dates.
    Use this to aggregate by month.
    """
    # Guard against empty dates
    if len(v) == 0:
        return []
    
    # Check if date values are pandas datetimes
    year_month = None
    if np.issubdtype(type(v[0]), str):
        #pull out the YYYY-MM part of the date string
        year_month = [x[:7] for x in v]
    else:
        #assume dates are datetime
        year_month = [d.strftime(r"%Y-%m") for d in v]
    return year_month
    

def get_month(dates):
    """
    Returns month, as a list of ints, for a given array of dates.
    """
    np_dates = to_np_datetimes64d(dates)
    answer = [(d.astype('datetime64[M]').astype(int) % 12 + 1) for d in np_dates]
    return answer


def get_dates(start_date: datetime, end_date=None, days=0, years=1, include_end_date=False, str_format=None):
    """
    Generates a list of daily datetime values from a given start date. The length 
    may be defined by an end_date, or a number of days, or a number of years. This 
    function may be useful for working with daily datasets and models.
    """
    if (days > 0):
        # great, we already have the number of days
        pass
    elif (end_date != None):
        # use end_date
        days = (end_date - start_date).days
        days = days + 1 if include_end_date else days
    else:
        # use years
        end_date = datetime(start_date.year + years, start_date.month, start_date.day,
            start_date.hour, start_date.minute, start_date.second, start_date.microsecond)
        days = (end_date - start_date).days
    date_list = [start_date + timedelta(days=x) for x in range(days)]
    if (str_format != None):
        date_list = [d.strftime(str_format) for d in date_list]
    return date_list

def get_wy_start_date(df:pd.DataFrame, wy_month=7):
    """
    Returns an appropriate water year start date based on data frame dates and the
    water year start month.

    Args:
        df (pd.DataFrame): Dataframe with date as index
        wy_month (int, optional): Water year start month. Defaults to 7.

    Returns:
        datetime: Water year start date.
    """
    first_date=df.index[0]
    first_day=int(first_date[8:10]) #0123-56-89
    first_month=int(first_date[5:7])
    first_year=int(first_date[0:4])

    if (first_month<wy_month):
        #If month is less than wy_month we can start wy this year
        start_month=wy_month
        start_day=1
        start_year=first_year
    elif (first_month==wy_month):
        #If month equal to wy_month check that data starts on first day of month and set year accordingly
        if (first_day>1):
            start_month=wy_month
            start_day=1
            start_year=first_year+1
        else:
            start_month=wy_month
            start_day=1
            start_year=first_year
    else:
        #If month is greater than wy_month we have to start wy next year
        start_month=wy_month
        start_day=1
        start_year=first_year+1
    
    return datetime(start_year,start_month,start_day)

def get_wy_end_date(df:pd.DataFrame, wy_month=7):
    """
    Returns an appropriate water year end date based on data frame dates and the
    water year start month.

    Args:
        df (pd.DataFrame): Dataframe with date as index
        wy_month (int, optional): Water year start month. Defaults to 7.

    Returns:
        datetime: Water year end date.
    """
    last_date=df.index[(len(df) - 1)]
    last_day=int(last_date[8:10]) #0123-56-89
    last_month=int(last_date[5:7])
    last_year=int(last_date[0:4])

    if (wy_month==1):
        wy_month_end=12
    else:
        wy_month_end=wy_month-1

    if wy_month_end in { 1, 3, 5, 7, 8, 10, 12 }:
        wy_day_end=31
    elif wy_month_end in { 4, 6, 9, 11 }:
        wy_day_end=30
    else:
        #Setting number of days in Feb to 28 - handle leap years at the end of this function
        wy_day_end=28

    if (last_month>wy_month_end):
        #If month is greater than wy_month_end we can start wy this year
        end_month=wy_month_end
        end_day=wy_day_end
        end_year=last_year
    elif (last_month==wy_month_end):
        #If month equal to wy_month_end check that data ends on last day of month and set year accordingly
        if (last_day<wy_day_end):
            end_month=wy_month_end
            end_day=wy_day_end
            end_year=last_year-1
        else:
            end_month=wy_month_end
            end_day=wy_day_end
            end_year=last_year
    else:
        #If month is less than wy_month_end we have to end wy last year
        end_month=wy_month_end
        end_day=wy_day_end
        end_year=last_year-1

    #This handles the February's that have 29 days
    if (end_month==2):
        end_day=(datetime(end_year,end_month+1,1) - timedelta(days=1)).day
    
    return datetime(end_year,end_month,end_day)


import pandas as pd
import numpy as np
from odhpy import utils
from datetime import datetime, timedelta
from typing import Union

class StorageLevelAssessment:
        
    def __init__(self, df: pd.Series, triggers: list, wy_month=7, allow_part_years=False) -> None:
        """
        Args:
            df (pd.Series): Daily timeseries of storage data with date as index.
            triggers (list): List of trigger thresholds to be assessed.
            wy_month (int, optional): Water year start month. Defaults to 7.
            allow_part_years (bool, optional): Allow part water years or only complete water years. Defaults to False.
        """

        if type(df) != pd.Series:
            raise Exception("Demand must be a single column of a dataframe (pd.Series)")   

        self.triggers = triggers
        self.wy_month = wy_month
        self.allow_part_years = allow_part_years
        self.df = df.copy(deep=True)

        # Calculate whether to include full WYs only
        if not allow_part_years:
            self.df = utils.crop_to_wy(self.df, wy_month)
        if (len(self.df) == 0):
            return np.nan    
        self.start_date=df.index[0]
        self.end_date=df.index[-1]

        # Run event algorithm on init. 
        self.events = {trigger: self.EventsBelowTriggerAlgorithm(trigger) for trigger in self.triggers}      

        # Get name of df Series
        self.columnname=self.df.name

        # Get count of WYs
        self.wy_count = self.df.groupby(utils.get_wy(self.df.index, self.wy_month)).sum().count()

    def AnnualDaysBelow(self):
        """Returns the total days at or below trigger threshold by WY.

        Returns:
            dict: Dictionary of annual timeseries grouped by trigger threshold.
        """

        dailytrigger = {trigger:pd.Series(np.where(self.df<=trigger,1,0),index=self.df.index) for trigger in self.triggers}
        annualdaysbelow = {trigger: x.groupby(utils.get_wy(x.index, self.wy_month)).sum() for trigger,x in dailytrigger.items()}
        return annualdaysbelow
    
    def AnnualDaysBelowSummary(self,trigger=None,annualdaysbelow=None):
        """Returns summary of total days at or below trigger threshold by WY.

        Args:
            trigger (any, optional): Optionally provide single trigger threshold to be assessed. Defaults to None.
            annualdaysbelow (dict, optional): Optionally provide output from AnnualDaysBelow, otherwise recalculate. Defaults to None.

        Returns:
            df: DataFrame of total days at or below threshold by WY, grouped by trigger threshold.
        """

        # If not provided, calculate AnnualDaysBelow
        if annualdaysbelow==None:
            annualdaysbelow = self.AnnualDaysBelow()

        # Output as DataFrame
        out_df=pd.DataFrame(annualdaysbelow)

        if trigger==None:
            return out_df
        else:
            return out_df[trigger]

    def NumberWaterYearsBelow(self,annualdaysbelow=None):
        """Returns total WYs with at least one day at or below trigger threshold.

        Args:
            annualdaysbelow (dict, optional): Optionally provide output from AnnualDaysBelow, otherwise recalculate. Defaults to None.

        Returns:
            dict: Dictionary of total years grouped by trigger threshold.
        """

        # If not provided, calculate AnnualDaysBelow        
        if annualdaysbelow==None:
            annualdaysbelow = self.AnnualDaysBelow()

        numberyears = {trigger: sum(1 if x > 0 else 0 for x in v) for trigger,v in annualdaysbelow.items()}
        return numberyears
    
    def PercentWaterYearsBelow(self,numberyears=None):
        """Returns percent of WYs with at least one day at or below trigger threshold.

        Args:
            numberyears (dict, optional): Optionally provide output from NumberWaterYearsBelow, otherwise recalculate. Defaults to None.

        Returns:
            dict: Dictionary of percent years grouped by trigger threshold.
        """

        # If not provided, calculate NumberWaterYearsBelow
        if numberyears==None:
            numberyears = self.NumberWaterYearsBelow()

        percyears = {trigger: x/self.wy_count for trigger,x in numberyears.items()}
        return percyears
    
    def EventsBelowTriggerAlgorithm(self,trigger):
        """Returns array of event lengths.

        Args:
            trigger (any): Trigger threshold against which daily data input is assessed.

        Returns:
            list: Array where each item represents the length of a single continuous event.
        """

        previous_ended=True
        length_counter=0
        event_counter=0
        output=[]

        # Determine last df index
        list_len = len(self.df)-1

        # For every daily value in df
        for index, x in enumerate(self.df):

            ## Storage less than or equal to trigger and currently in event
            # Add to count
            if x <= trigger and previous_ended==False:
                length_counter=length_counter+1

            ## Storage less than or equal to trigger and not in an event
            # Append current length count to output array (if not in first event)
            # Start new event
            # Add to count
            if x <= trigger and previous_ended:
                # If not first event
                if event_counter > 0:
                    output.append(length_counter)
                    length_counter=0
                previous_ended=False
                length_counter=length_counter+1
                event_counter=event_counter+1
            
            ## Storage greater than trigger
            # End current event
            if x > trigger:
                previous_ended=True

            # If at last day, append current length count to output array
            if index==list_len:
                if event_counter > 0:
                    output.append(length_counter)
                    length_counter=0

        return output

    def EventsBelowTrigger(self,length=1):
        """Returns event length array for each trigger threshold.

        Args:
            length (int, optional): Optional minimum event length to return. Defaults to 1.

        Returns:
            dict: Dictionary of event length arrays, grouped by trigger threshold.
        """

        trunc_events = {k:[i for i in x if i>=length] for k,x in self.events.items()}
        return trunc_events
        
    def EventsBelowTriggerCount(self,length=1):
        """Returns count of events for each trigger threshold

        Args:
            length (int, optional): Optional minimum event length to count. Defaults to 1.

        Returns:
            dict: Dictionary of event counts, grouped by trigger threshold.
        """
        
        output = {k:sum(i >= length for i in x) for k,x in self.events.items()}
        return output
    
    def EventsBelowTriggerMax(self):
        """Returns max event length for each trigger threshold

        Returns:
            dict: Dictionary of event counts, grouped by trigger threshold.
        """
                
        output = {k:max(x) if len(x)>0 else np.nan for k,x in self.events.items()}
        return output
    
    def Summary(self,trigger=None):
        """Returns table summary of key storage level assessment outputs.

        Args:
            trigger (any, optional): Optionally provide single trigger threshold to be assessed. Defaults to None.

        Returns:
            df: Dataframe summary
        """

        out_df = pd.DataFrame()
        temp_numberyears=self.NumberWaterYearsBelow()
        out_df['Column name']={trigger: self.columnname for trigger in self.triggers}
        out_df['Start date']={trigger: self.start_date for trigger in self.triggers}
        out_df['End date']={trigger: self.end_date for trigger in self.triggers}
        out_df['Number water years with at least 1 day at or below level']=temp_numberyears
        out_df['Percentage water years with at least 1 day at or below level']=self.PercentWaterYearsBelow(temp_numberyears)
        out_df['Number of events at or below trigger (>=1day)']=self.EventsBelowTriggerCount()
        out_df['Number of events at or below trigger (>=7days)']=self.EventsBelowTriggerCount(7)
        out_df['Number of events at or below trigger (>=30days)']=self.EventsBelowTriggerCount(30)
        out_df['Number of events at or below trigger (>=91days)']=self.EventsBelowTriggerCount(91)
        out_df['Number of events at or below trigger (>=183days)']=self.EventsBelowTriggerCount(183)
        out_df['Number of events at or below trigger (>=365days)']=self.EventsBelowTriggerCount(365)
        out_df['Longest period at or below trigger (days)']=self.EventsBelowTriggerMax()

        # If trigger is provided, subset those outputs
        if trigger==None:
            return out_df
        else:
            return out_df.loc[trigger]



import calendar

def from_pattern(dates, daily_pattern=[1], monthly_pattern=[], annual_pattern=[]):
    """Generates a list of daily values from either a daily, montly, or annual pattern.
    The behaviour depends on what type of pattern is supplied. 
    Annual pattern => annual totals are disaggregated to daily by dividing by 365 or 366 as appropriate.
    Monthly pattern => 
        if 1 value is provided, it is used for every month and disaggregated to daily by dividing by the days in the month;
        if 12 values are provided, each month has its own total and which is disaggregated to daily by dividing by the days in the month.
    Daily pattern => 
        if 1 value is provided, it is used for every day;
        if 7 values are provided, they are taken to represent a week and are repeated accordingly;
        if 366 values are provided, they will be taken to represent days in the calendar year and the last value ignored in non-leap-years.

    Args:
        dates (_type_): _description_
        daily_pattern (list, optional): _description_. Defaults to [1].
        monthly_pattern (list, optional): _description_. Defaults to [].
        annual_pattern (list, optional): _description_. Defaults to [].

    Raises:
        Exception: _description_
        Exception: _description_

    Returns:
        _type_: _description_
    """
    # Guard
    if len(dates) == 0:
        return []
    # Generate data
    if len(annual_pattern) > 0:
        # Annual totals disaggregated to daily
        pattern_len = len(annual_pattern)
        year0 = dates[0].year
        return [annual_pattern[(d.year - year0) % pattern_len]/(366 if calendar.isleap(d.year) else 365) for d in dates]
    elif len(monthly_pattern) > 1:
        # Monthly pattern disaggregated to daily
        if (len(monthly_pattern) == 1):
            return [monthly_pattern[0]/calendar.monthrange(d.year,d.month)[1] for d in dates]
        if (len(monthly_pattern) == 12):
            # Pattern assumed to start in January
            return [monthly_pattern[d.month - 1]/calendar.monthrange(d.year,d.month)[1] for d in dates]
        else:
            raise Exception("Monthly pattern must have 1 or 12 values.")
    else:
        # Daily pattern
        if len(daily_pattern) == 1:
            value = daily_pattern[0]
            return [value for _ in range(len(dates))]
        elif len(daily_pattern) == 7:
            # Pattern assumed to start on Monday
            return [daily_pattern[d.weekday()] for d in dates]
        elif len(daily_pattern) == 366:
            # Pattern assumed to start on Jan 1 (non-leapyears will skip the last value)
            return [daily_pattern[d.timetuple().tm_yday - 1] for d in dates]
        else:
            raise Exception("Daily pattern must have 1, 7 or 366 values.")


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

exec(open('src/odhpy/version.py').read())

setuptools.setup(
    name="odhpy",
    version=__version__,
    python_requires=">=3.9",
    author="Chas Egan",
    author_email="chas@odhydrology.com",
    description="A collection of splashings to master that which has no form and count that which is uncountable.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://bitbucket.org/odhydrology/odhpy.git",
    package_dir={'': 'src'},
    packages=setuptools.find_packages('src'),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'altair>=5.0.1',
        'folium>=0.14',
        'matplotlib>=3.8.3',
        'numpy>=1.26.4',
        'pandas>=2.2.0',
        'plotly>=5.18.0',
    ],    
)


import pandas as pd
from datetime import datetime, timedelta
from odhpy import utils


def join_on_dates(df1, df2, assert_format_standards=True):
    """_summary_
    Assumes df1 and df2 both have datetime indexes with values that are daily,
    and sequentially ascending.

    Args:
        df1 (_type_): _description_
        df2 (_type_): _description_

    Returns:
        _type_: _description_
    """
    if assert_format_standards:
        utils.assert_df_format_standards(df1)
        utils.assert_df_format_standards(df2)
    min_datetime = datetime.strptime(min(min(df1.index),min(df2.index)), r"%Y-%m-%d")
    max_datetime = datetime.strptime(max(max(df1.index),max(df2.index)), r"%Y-%m-%d")
    ans_df = pd.DataFrame()
    ans_df["Date"] = utils.get_dates(min_datetime, max_datetime, str_format=r"%Y-%m-%d")
    ans_df.set_index("Date", inplace=True)
    ans_df = ans_df.join(df1, how='left')
    ans_df = ans_df.join(df2, how='left', lsuffix='_left', rsuffix='_right')
    if assert_format_standards:
        utils.assert_df_format_standards(ans_df)
    return ans_df



def get_exceedence(obs_df, mod_df, plotting_position="cunnane") -> pd.DataFrame:
    """_summary_

    Args:
        obs_df (_type_): _description_
        mod_df (_type_): _description_
        plotting_position (str, optional): _description_. Defaults to "cunnane". Other supported values: "weibull", "gringorten". See https://glossary.ametsoc.org/wiki/Plotting_position

    Raises:
        Exception: _description_

    Returns:
        _type_: _description_
    """
    df = obs_df.join(mod_df, how='inner')
    df = df.dropna()
    df.columns = ['x', 'y']
    df.x = df.x.sort_values(ascending=False).values
    df.y = df.y.sort_values(ascending=False).values
    n = len(df)
    df.index = get_exceedence_plotting_position(n, plotting_position)
    return df



def get_exceedence_plotting_position(length, plotting_position="cunnane"):
    """Get plotting position values for a given length.

    Args:
        plotting_position (str, optional): _description_. Defaults to "cunnane". Other supported values: "weibull", "gringorten". See https://glossary.ametsoc.org/wiki/Plotting_position

    Raises:
        Exception: If plotting position is not supported.

    Returns:
        _type_: _description_
    """
    
    n = length

    index_starting_at_one = [i + 1 for i in range(n)]
    if plotting_position == "cunnane":
        plotting_points = [100 * (r - 0.4)/(n + 0.2) for r in index_starting_at_one]
    elif plotting_position == "weibull":
        plotting_points = [100 * (r/(n + 1)) for r in index_starting_at_one]
    elif plotting_position == "gringorten":
        plotting_points = [100 * (r - 0.44)/(n + 0.12) for r in index_starting_at_one]
    else:
        raise Exception(f"Plotting position not supported: {plotting_position}")
    return plotting_points